"""
This module provides the main EchoStateNetwork instance, which can be trained to forecast time series
"""


from dataclasses import dataclass, field
from typing import Optional, Callable, Union
import warnings

import numpy as np

from .time_series import TimeSeries


class InstabilityWarning(Warning):

    """
    Warning indicating some sort of numerical instability
    """

    pass


@dataclass
class EchoStateNetwork:

    """
    ESN class for making prediction
    """

    reservoir_dimensionality: int
    """dimensionality of reservoir, should be much larger than input dimension"""
    input_dimensionality: int
    """expected dimensionality of input time series"""
    w_in: Optional[np.typing.NDArray] = None
    """matrix defining input -> reservoir mapping, autogenerated if not specified"""
    regularization_parameter: float = 1.0e-10
    """regularization parameter for fit, should be small, defaults to 1.0e-10"""
    generator: Optional[np.random.Generator] = None
    """random generator, auto generated if not specified"""
    activation_function: Callable[[np.typing.NDArray], np.typing.NDArray] = np.tanh
    """activation function squishing reservoir states, defaults to hyperbolic tangent"""
    w_out: np.typing.NDArray = field(init=False)
    """matrix defining reservoir -> output mapping; trainable and should not be set at initialization"""
    feedback_loop_guess: Union[float, np.typing.NDArray] = field(init=False)
    """starting warmup for prediction; trainable and should not be set at initialization"""
    timestep: float = field(init=False)
    """detected timestep of inputted TimeSeries; trainable and should not be set at initialization"""
    final_time: float = field(init=False)
    """final time of inputted TimeSeries; trainable and should not be set at initialization"""

    def __post_init__(self):

        """
        Need to initialize a generator and a W_in.
        The generator attribute here is responsible for random sampling
        """

        if not self.generator:
            self.generator = np.random.default_rng(seed=0)

        if self.w_in is None:
            self.w_in = self.generator.uniform(
                low=-0.5,
                high=0.5,
                size=(self.reservoir_dimensionality, self.input_dimensionality)
            )

    def train(self, input_time_series: TimeSeries, pinv: bool = False, max_condition_number: float = 10.0):

        """
        Training step, solving argmin_W ||Y - WX||^2 + λ||W||^2

        Args:
            input_time_series (TimeSeries):
                The input TimeSeries instance to train on

            pinv (bool):
                Uses the Moore-Penrose inverse in the training if true. Defaults to false.

            max_condition_number (float):
                Maximum condition number of the training data.
                A larger condition number will induce an InstabilityWarning
        """

        # assert that all timesteps are constant
        self.timestep = input_time_series.timestep
        self.final_time = input_time_series.times[-1]

        time_series_array = input_time_series.dependent_variable
        independent_variables = self.activation_function(self.w_in @ time_series_array[:-1, :].T).T
        dependent_variables = time_series_array[1:]

        regularization = self.regularization_parameter * np.eye(independent_variables.shape[1])

        if pinv:
            w_out_transpose = np.linalg.pinv(independent_variables) @ dependent_variables
        else:
            x = independent_variables.T @ independent_variables + regularization

            # conditional number 
            cond_num = np.linalg.cond(x)
            if cond_num > max_condition_number:
                warnings.warn(
                    f"Condition Number {cond_num:.2E} is greater than {max_condition_number}. "
                    f"consider passing pinv = True in {self.__class__.__name__}.train() "
                    f"or increasing {self.__class__.__name__}.regularization_parameter",
                    InstabilityWarning
                )

            w_out_transpose = np.linalg.solve(x, independent_variables.T @ dependent_variables)
            
        self.w_out = w_out_transpose.T
        self.feedback_loop_guess = time_series_array[-1]

    def predict(self, horizon: int) -> TimeSeries:

        """
        Prediction step, using the previous training to forecast into the future

        Args:
            horizon (int): The number of steps to predict into the future

        Returns:
            TimeSeries: The resulting predicted time series
        """

        if self.w_out is None or self.w_in is None:
            raise ValueError("need to train before predicting")

        # initialize predictions and reservoir states to populate later
        predictions = np.zeros((horizon, self.input_dimensionality))

        # perform feedback loop
        for i in range(horizon):
            if i == 0:
                predictions[i, :] = self.w_out @ self.activation_function(self.w_in @ self.feedback_loop_guess)
                continue
            predictions[i, :] = self.w_out @ self.activation_function(self.w_in @ predictions[i - 1, :])

        return TimeSeries(
            dependent_variable=predictions,
            times=self.final_time + self.timestep + np.linspace(0, horizon * self.timestep, horizon)
        )



from typing import IO, Union
from pathlib import Path
import warnings

from dataclasses import dataclass
import numpy as np

from numpy.typing import NDArray


class ShapeChangedWarning(Warning):

    """
    warn the user that TimeSeries.__post_init__ changed the shape of the dependent variable
    """


@dataclass
class TimeSeries:

    dependent_variable: NDArray
    times: NDArray

    def __post_init__(self):

        timesteps = np.diff(self.times)
        if not np.isclose(np.std(timesteps), 0.0):
            raise ValueError("TimeSeries.times must be uniformly spaced")
        if np.isclose(np.mean(timesteps), 0.0):
            raise ValueError("TimeSeries.times must have a timestep greater than zero")

        if len(self.dependent_variable.shape) == 1:
            num_steps = len(self.dependent_variable)
            self.dependent_variable = self.dependent_variable.reshape(num_steps, 1)
            warnings.warn(
                f"TimeSeries.dependent_variable should have shape (number of steps, dimensionality). "
                f"The shape has been changed from {(num_steps,)} to {self.dependent_variable.shape}"
            )

    def save(self, fp: Union[IO, str, Path], header: str = "", delimiter=","):

        """
        Method to save a time series

        Args:
            fp (Union[IO, str, Path]):
                The file-like object, path name, or Path in which to save the TimeSeries instance

            header (str):
                An optional header. Defaults to the empty string

            delimiter (str):
                The delimiting character in the save file. Defaults to a comma

        """
        np.savetxt(
            fp,
            np.vstack((self.times, self.dependent_variable.T)).T,
            delimiter=delimiter,
            header=header,
            comments=""
        )

    @property
    def num_dims(self) -> int:

        """
        The dimensionality of the time series

        Returns:
            int: The dimensionality of the time series
        """

        return self.dependent_variable.shape[1]

    @classmethod
    def from_csv(cls, fp: Union[IO, str, Path], time_index: int = 0):

        """
        Method for loading in a TimeSeries instance from a comma-separated value (csv) file

        Args:
            fp (Union[IO, str, Path]):
                The file-like object, path name, or Path in which to read

            time_index (int):
                The column index corresponding to the time column. Defaults to 0

        Returns:
            TimeSeries: A TimeSeries instance populated by data from the csv file
        """

        data = np.loadtxt(fp, delimiter=",")

        return cls(
            dependent_variable=np.delete(data, obj=time_index, axis=1),
            times=data[:, time_index]
        )

    @property
    def timestep(self) -> float:

        """
        The physical timestep of the time series

        Returns:
            int: The physical timestep of the time series
        """

        return self.times[1] - self.times[0]

    def __eq__(self, other) -> bool:
        return bool(np.all(self.times == other.times) and np.all(
            np.isclose(self.dependent_variable, other.dependent_variable)
        ))

    def __getitem__(self, key):

        return TimeSeries(self.dependent_variable[key], self.times[key])

    def __setitem__(self, key, value):

        if not isinstance(value, TimeSeries):
            raise TypeError("Value must be a TimeSeries object")
        if isinstance(key, slice) and key.stop > len(self.dependent_variable):
            raise ValueError("Slice stop index out of range")
        if isinstance(key, int) and key >= len(self.dependent_variable):
            raise ValueError("Index out of range")

        self.dependent_variable[key] = value.dependent_variable
        self.times[key] = value.times

    def __add__(self, other):

        if not len(self.times) == len(other.times):
            raise ValueError("can only add TimeSeries instances that have the same number of timesteps")

        if not np.all(self.times == other.times):
            raise ValueError("can only add TimeSeries instances that span the same times")

        return TimeSeries(
            dependent_variable=self.dependent_variable + other.dependent_variable,
            times=self.times
        )

    def __sub__(self, other):

        if not len(self.times) == len(other.times):
            raise ValueError("can only subtract TimeSeries instances that have the same number of timesteps")

        if not np.all(self.times == other.times):
            raise ValueError("can only subtract TimeSeries instances that span the same times")

        return TimeSeries(
            dependent_variable=self.dependent_variable - other.dependent_variable,
            times=self.times
        )

    def __rshift__(self, other):

        if self.times[-1] >= other.times[0]:
            raise ValueError("can only concatenate TimeSeries instances with non-overlapping time values")

        return TimeSeries(
            dependent_variable=np.vstack((self.dependent_variable, other.dependent_variable)),
            times=np.hstack((self.times, other.times))
        )



__version__ = "0.0.18"
__authors__ = [
    "Jacob Jeffries",
    "Ameen Mahmood",
    "Avik Thumati",
    "Samuel Josephs"
]
__author_emails__ = [
    "jacob.jeffries@chicagoclubteam.org",
    "ameen.mahmood@chicagoclubteam.org",
    "avik.thumati@chicagoclubteam.org",
    "samuel.joseph@chicagoclubteam.org"
]
__url__ = "https://github.com/Chicago-Club-Management-Company/aqua-blue"

from .reservoir import EchoStateNetwork as EchoStateNetwork
from .reservoir import InstabilityWarning as InstabilityWarning
from .time_series import TimeSeries as TimeSeries
from .time_series import ShapeChangedWarning as ShapeChangedWarning



"""
This module provides simple utilities for processing TimeSeries instances
"""


from dataclasses import dataclass, field

from numpy.typing import NDArray

from .time_series import TimeSeries


@dataclass
class Normalizer:

    """
    Normalizer class to normalize a time series to have zero mean and unit variance
    """

    means: NDArray = field(init=False)
    standard_deviations: NDArray = field(init=False)

    def normalize(self, time_series: TimeSeries) -> TimeSeries:

        """
        Normalize the given TimeSeries instance

        Args:
            time_series (TimeSeries): Time series to normalize

        Returns:
            TimeSeries: The normalized time series

        Raises:
            ValueError: If the normalizer has already been used
        """

        if hasattr(self, "means") or hasattr(self, "standard_deviations"):
            raise ValueError("You can only use the Normalizer once. Create a new instance to normalize again")

        arr = time_series.dependent_variable
        self.means = arr.mean(axis=0)
        self.standard_deviations = arr.std(axis=0)

        arr = arr - self.means
        arr = arr / self.standard_deviations

        return TimeSeries(
            dependent_variable=arr,
            times=time_series.times
        )

    def denormalize(self, time_series: TimeSeries) -> TimeSeries:

        """
        Denormalize the given TimeSeries instance. Means and standard deviations are grabbed
        from a previous normalization

        Args:
            time_series (TimeSeries): Time series to denormalize

        Returns:
            TimeSeries: The denormalized time series

        Raises:
            ValueError: If the normalizer has not yet been used
        """

        if not hasattr(self, "means") or not hasattr(self, "standard_deviations"):
            raise ValueError("You can only denormalize after normalizing a time series")

        arr = time_series.dependent_variable
        arr = arr * self.standard_deviations
        arr = arr + self.means

        return TimeSeries(dependent_variable=arr, times=time_series.times)
